"""OME-Zarr HCS objects models."""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ngio.images import OmeZarrContainer


class OmeZarrPlate:
    """Placeholder for the OME-Zarr image object."""

    def __init__(self, *args, **kwargs):
        """Initialize the OME-Zarr plate."""
        raise NotImplementedError

    def wells(self) -> list[str]:
        """Return the wells."""
        raise NotImplementedError

    def columns(self) -> list[str]:
        """Return the number of columns."""
        raise NotImplementedError

    def rows(self) -> list[str]:
        """Return the number of rows."""
        raise NotImplementedError

    def get_omezarr_well(self, *args, **kwargs) -> "OmeZarrWell":
        """Return the OME-Zarr well."""
        raise NotImplementedError

    def get_omezarr_image(self, *args, **kwargs) -> "OmeZarrContainer":
        """Return the OME-Zarr image."""
        raise NotImplementedError


class OmeZarrWell:
    """Placeholder for the Image object."""

    def __init__(self, *args, **kwargs):
        """Initialize the OME-Zarr well."""
        raise NotImplementedError

    def acquisitions(self) -> list[str]:
        """Return the acquisition."""
        raise NotImplementedError

    def get_ome_zarr_image(self, *args, **kwargs) -> "OmeZarrContainer":
        """Return the OME-Zarr image."""
        raise NotImplementedError


def open_omezarr_plate(*args, **kwargs):
    """Open an OME-Zarr plate."""
    return OmeZarrPlate(*args, **kwargs)


def open_omezarr_well(*args, **kwargs):
    """Open an OME-Zarr well."""
    return OmeZarrWell(*args, **kwargs)



from pathlib import Path

from ngio.tables.tables_container import TablesContainer, open_tables_container


def test_table_container(tmp_path: Path):
    table_group = open_tables_container(tmp_path / "test.zarr")
    # TODO test the table group
    assert isinstance(table_group, TablesContainer)



from typing import Any, Generic, TypeVar

from pydantic import ValidationError

from ngio.ome_zarr_meta._generic_handlers import (
    ImageMetaHandler,
    LabelMetaHandler,
)
from ngio.ome_zarr_meta.ngio_specs import AxesSetup
from ngio.ome_zarr_meta.v04 import V04ImageMetaHandler, V04LabelMetaHandler
from ngio.utils import (
    AccessModeLiteral,
    NgioValidationError,
    NgioValueError,
    StoreOrGroup,
    ZarrGroupHandler,
)

_Image_or_Label_Plugin = TypeVar(
    "_Image_or_Label_Plugin", ImageMetaHandler, LabelMetaHandler
)


class _ImplementedMetaHandlers(Generic[_Image_or_Label_Plugin]):
    """This class is a singleton that manages the available image handler plugins."""

    _instance = None
    _implemented_handlers: dict[str, Any]

    def __new__(cls):
        """Create a new instance of the class if it does not exist."""
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._implemented_handlers = {}
        return cls._instance

    def available_handlers(self) -> list[str]:
        """Get the available image handler versions.

        The versions are returned in descending order.
            such that the latest version is the first in the list and the fist to be
            checked.
        """
        return list(reversed(self._implemented_handlers.keys()))

    def find_meta_handler(
        self,
        group_handler: ZarrGroupHandler,
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = True,
    ) -> _Image_or_Label_Plugin:
        """Try to get a handler for the given store based on the metadata version."""
        _errors = {}

        for version, handler in reversed(self._implemented_handlers.items()):
            handler = handler(
                group_handler=group_handler,
                axes_setup=axes_setup,
                allow_non_canonical_axes=allow_non_canonical_axes,
                strict_canonical_order=strict_canonical_order,
            )
            meta = handler.safe_load_meta()
            if isinstance(meta, ValidationError):
                _errors[version] = meta
                continue
            return handler

        raise NgioValidationError(
            f"Could not load OME-Zarr metadata from any known version. "
            f"Errors: {_errors}"
        )

    def get_handler(
        self,
        version: str,
        group_handler: ZarrGroupHandler,
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = True,
    ) -> _Image_or_Label_Plugin:
        """Get a handler for a specific version."""
        if version not in self._implemented_handlers:
            raise NgioValueError(f"Image handler for version {version} does not exist.")
        return self._implemented_handlers[version](
            group_handler=group_handler,
            axes_setup=axes_setup,
            allow_non_canonical_axes=allow_non_canonical_axes,
            strict_canonical_order=strict_canonical_order,
        )

    def add_handler(
        self, key: str, handler: type[_Image_or_Label_Plugin], overwrite: bool = False
    ):
        """Register a new handler."""
        if key in self._implemented_handlers and not overwrite:
            raise NgioValueError(f"Image handler for version {key} already exists.")
        self._implemented_handlers[key] = handler


class ImplementedImageMetaHandlers(_ImplementedMetaHandlers[ImageMetaHandler]):
    def __init__(self):
        super().__init__()


ImplementedImageMetaHandlers().add_handler("0.4", V04ImageMetaHandler)


class ImplementedLabelMetaHandlers(_ImplementedMetaHandlers[LabelMetaHandler]):
    def __init__(self):
        super().__init__()


ImplementedLabelMetaHandlers().add_handler("0.4", V04LabelMetaHandler)


def open_image_meta_handler(
    store: StoreOrGroup,
    cache: bool = False,
    mode: AccessModeLiteral = "a",
    axes_setup: AxesSetup | None = None,
    allow_non_canonical_axes: bool = False,
    strict_canonical_order: bool = True,
) -> ImageMetaHandler:
    """Open the metadata of an OME-Zarr image.

    Args:
        store: The Zarr store or group where the model is
        cache: Whether to cache the metadata.
        mode: The store access mode.
        axes_setup: The axes setup. This is used to map axes with
            a non-canonical name to a canonical name.
        allow_non_canonical_axes: Whether to allow non-canonical axes.
        strict_canonical_order: Whether to enforce strict canonical order.
    """
    zarr_group_handler = ZarrGroupHandler(store, cache, mode)
    return ImplementedImageMetaHandlers().find_meta_handler(
        zarr_group_handler,
        axes_setup=axes_setup,
        allow_non_canonical_axes=allow_non_canonical_axes,
        strict_canonical_order=strict_canonical_order,
    )



from pathlib import Path

import pytest

from ngio.utils import download_ome_zarr_dataset


@pytest.fixture(scope="function")
def ome_zarr_image_v04_path(tmp_path: Path) -> Path:
    tmp_path = tmp_path / "test_datasets_v04"
    return Path(download_ome_zarr_dataset("CardiomyocyteTiny", download_dir=tmp_path))



from pathlib import Path

from ngio.ome_zarr_meta import NgioImageMeta, open_image_meta_handler


def test_get_image_handler(ome_zarr_image_v04_path: Path):
    # TODO this is a placeholder test
    # The pooch cache is giving us trouble here
    ome_zarr_image_v04_path = ome_zarr_image_v04_path / "B" / "03" / "0"
    handler = open_image_meta_handler(ome_zarr_image_v04_path, cache=True, mode="a")
    meta = handler.safe_load_meta()
    assert isinstance(meta, NgioImageMeta)
    handler.write_meta(meta)



"""Fractal internal module for dataset metadata handling."""

import numpy as np

from ngio.ome_zarr_meta.ngio_specs import SpaceUnits, TimeUnits

################################################################################################
#
# PixelSize model
# The PixelSize model is used to store the pixel size in 3D space.
# The model does not store scaling factors and units for other axes.
#
#################################################################################################


def _validate_type(value: float, name: str) -> float:
    """Check the type of the value."""
    if not isinstance(value, int | float):
        raise TypeError(f"{name} must be a number.")
    return float(value)


class PixelSize:
    """PixelSize class to store the pixel size in 3D space."""

    def __init__(
        self,
        x: float,
        y: float,
        z: float,
        t: float = 0,
        space_unit: SpaceUnits = SpaceUnits.micrometer,
        time_unit: TimeUnits | None = TimeUnits.s,
    ):
        """Initialize the pixel size."""
        self.x = _validate_type(x, "x")
        self.y = _validate_type(y, "y")
        self.z = _validate_type(z, "z")
        self.t = _validate_type(t, "t")

        if not isinstance(space_unit, SpaceUnits):
            raise TypeError("space_unit must be of type SpaceUnits.")
        self.space_unit = space_unit

        if time_unit is not None and not isinstance(time_unit, TimeUnits):
            raise TypeError("time_unit must be of type TimeUnits.")
        self.time_unit = time_unit

    def __repr__(self) -> str:
        """Return a string representation of the pixel size."""
        return f"PixelSize(x={self.x}, y={self.y}, z={self.z}, t={self.t})"

    def as_dict(self) -> dict:
        """Return the pixel size as a dictionary."""
        return {"t": self.t, "z": self.z, "y": self.y, "x": self.x}

    @property
    def zyx(self) -> tuple[float, float, float]:
        """Return the voxel size in z, y, x order."""
        return self.z, self.y, self.x

    @property
    def yx(self) -> tuple[float, float]:
        """Return the xy plane pixel size in y, x order."""
        return self.y, self.x

    @property
    def voxel_volume(self) -> float:
        """Return the volume of a voxel."""
        return self.y * self.x * self.z

    @property
    def xy_plane_area(self) -> float:
        """Return the area of the xy plane."""
        return self.y * self.x

    @property
    def time_spacing(self) -> float | None:
        """Return the time spacing."""
        return self.t

    def distance(self, other: "PixelSize") -> float:
        """Return the distance between two pixel sizes in 3D space."""
        return float(np.linalg.norm(np.array(self.zyx) - np.array(other.zyx)))



"""Implementation of the ROI Table class.

This class follows the roi_table specification at:
https://fractal-analytics-platform.github.io/fractal-tasks-core/tables/
"""

from collections.abc import Iterable
from typing import Literal

import pandas as pd
from pydantic import BaseModel

from ngio.common import WorldCooROI
from ngio.tables._validators import validate_columns
from ngio.tables.backends import ImplementedTableBackends
from ngio.utils import NgioValueError, ZarrGroupHandler

REQUIRED_COLUMNS = [
    "x_micrometer",
    "y_micrometer",
    "z_micrometer",
    "len_x_micrometer",
    "len_y_micrometer",
    "len_z_micrometer",
]


ORIGIN_COLUMNS = [
    "x_micrometer_original",
    "y_micrometer_original",
]

TRANSLATION_COLUMNS = ["translation_x", "translation_y", "translation_z"]

OPTIONAL_COLUMNS = ORIGIN_COLUMNS + TRANSLATION_COLUMNS


def _dataframe_to_rois(dataframe: pd.DataFrame) -> dict[str, WorldCooROI]:
    """Convert a DataFrame to a WorldCooROI object."""
    rois = {}
    for key, row in dataframe.iterrows():
        # check if optional columns are present
        origin = {col: row.get(col, None) for col in ORIGIN_COLUMNS}
        origin = dict(filter(lambda x: x[1] is not None, origin.items()))
        translation = {col: row.get(col, None) for col in TRANSLATION_COLUMNS}
        translation = dict(filter(lambda x: x[1] is not None, translation.items()))

        roi = WorldCooROI(
            name=str(key),
            x=row["x_micrometer"],
            y=row["y_micrometer"],
            z=row["z_micrometer"],
            x_length=row["len_x_micrometer"],
            y_length=row["len_y_micrometer"],
            z_length=row["len_z_micrometer"],
            unit="micrometer",  # type: ignore
            **origin,
            **translation,
        )
        rois[roi.name] = roi
    return rois


def _rois_to_dataframe(rois: dict[str, WorldCooROI], index_key: str) -> pd.DataFrame:
    """Convert a list of WorldCooROI objects to a DataFrame."""
    data = []
    for roi in rois.values():
        row = {
            index_key: roi.name,
            "x_micrometer": roi.x,
            "y_micrometer": roi.y,
            "z_micrometer": roi.z,
            "len_x_micrometer": roi.x_length,
            "len_y_micrometer": roi.y_length,
            "len_z_micrometer": roi.z_length,
        }

        extra = roi.model_extra or {}
        for col in ORIGIN_COLUMNS:
            if col in extra:
                row[col] = extra[col]

        for col in TRANSLATION_COLUMNS:
            if col in extra:
                row[col] = extra[col]
        data.append(row)
    dataframe = pd.DataFrame(data)
    dataframe = dataframe.set_index(index_key)
    return dataframe


class ROITableV1Meta(BaseModel):
    """Metadata for the ROI table."""

    fractal_table_version: Literal["1"] = "1"
    type: Literal["roi_table"] = "roi_table"
    backend: str | None = None


class RoiTableV1:
    """Class to handle fractal ROI tables.

    To know more about the ROI table format, please refer to the
    specification at:
    https://fractal-analytics-platform.github.io/fractal-tasks-core/tables/
    """

    def __init__(self, rois: Iterable[WorldCooROI] | None = None) -> None:
        """Create a new ROI table."""
        self._meta = ROITableV1Meta()
        self._table_backend = None

        self._rois = {}
        if rois is not None:
            self.add(rois)

    @staticmethod
    def type() -> Literal["roi_table"]:
        """Return the type of the table."""
        return "roi_table"

    @staticmethod
    def version() -> Literal["1"]:
        """Return the version of the fractal table."""
        return "1"

    @property
    def backend_name(self) -> str | None:
        """Return the name of the backend."""
        if self._table_backend is None:
            return None
        return self._table_backend.backend_name()

    @classmethod
    def _from_handler(
        cls, handler: ZarrGroupHandler, backend_name: str | None = None
    ) -> "RoiTableV1":
        """Create a new ROI table from a Zarr store."""
        meta = ROITableV1Meta(**handler.load_attrs())

        if backend_name is None:
            backend = ImplementedTableBackends().get_backend(
                backend_name=meta.backend,
                group_handler=handler,
                index_key="FieldIndex",
                index_type="str",
            )
        else:
            backend = ImplementedTableBackends().get_backend(
                backend_name=backend_name,
                group_handler=handler,
                index_key="FieldIndex",
                index_type="str",
            )
            meta.backend = backend_name

        if not backend.implements_dataframe:
            raise NgioValueError(
                "The backend does not implement the dataframe protocol."
            )

        table = cls()
        table._meta = meta
        table._table_backend = backend

        dataframe = backend.load_as_dataframe()
        dataframe = validate_columns(
            dataframe,
            required_columns=REQUIRED_COLUMNS,
            optional_columns=OPTIONAL_COLUMNS,
        )
        table._rois = _dataframe_to_rois(dataframe)
        return table

    def _set_backend(
        self,
        handler: ZarrGroupHandler,
        backend_name: str | None = None,
    ) -> None:
        """Set the backend of the table."""
        backend = ImplementedTableBackends().get_backend(
            backend_name=backend_name,
            group_handler=handler,
            index_key="FieldIndex",
            index_type="str",
        )
        self._meta.backend = backend_name
        self._table_backend = backend

    def rois(self) -> list[WorldCooROI]:
        """List all ROIs in the table."""
        return list(self._rois.values())

    def get(self, roi_name: str) -> WorldCooROI:
        """Get an ROI from the table."""
        if roi_name not in self._rois:
            raise NgioValueError(f"ROI {roi_name} not found in the table.")
        return self._rois[roi_name]

    def add(self, roi: WorldCooROI | Iterable[WorldCooROI]) -> None:
        """Append ROIs to the current table."""
        if isinstance(roi, WorldCooROI):
            roi = [roi]

        for _roi in roi:
            if _roi.name in self._rois:
                raise NgioValueError(f"ROI {_roi.name} already exists in the table.")
            self._rois[_roi.name] = _roi

    def consolidate(self) -> None:
        """Write the current state of the table to the Zarr file."""
        if self._table_backend is None:
            raise NgioValueError(
                "No backend set for the table. "
                "Please add the table to a OME-Zarr Image before calling consolidate."
            )

        dataframe = _rois_to_dataframe(self._rois, index_key="FieldIndex")
        dataframe = validate_columns(
            dataframe,
            required_columns=REQUIRED_COLUMNS,
            optional_columns=OPTIONAL_COLUMNS,
        )
        self._table_backend.write_from_dataframe(
            dataframe, metadata=self._meta.model_dump(exclude_none=True)
        )



"""Generic class to handle Image-like data in a OME-NGFF file."""

from collections.abc import Collection
from typing import Literal

from dask import array as da

from ngio.common import Dimensions
from ngio.images.abstract_image import AbstractImage, consolidate_image
from ngio.images.create import _create_empty_image
from ngio.ome_zarr_meta import (
    ImageMetaHandler,
    ImplementedImageMetaHandlers,
    NgioImageMeta,
    PixelSize,
)
from ngio.ome_zarr_meta.ngio_specs import Channel, ChannelsMeta, ChannelVisualisation
from ngio.utils import (
    NgioValidationError,
    StoreOrGroup,
    ZarrGroupHandler,
)


def _check_channel_meta(meta: NgioImageMeta, dimension: Dimensions) -> ChannelsMeta:
    """Check the channel metadata."""
    c_dim = dimension.get("c", strict=False)
    c_dim = 1 if c_dim is None else c_dim

    if meta.channels_meta is None:
        return ChannelsMeta.default_init(labels=c_dim)

    if len(meta.channels) != c_dim:
        raise NgioValidationError(
            "The number of channels does not match the image. "
            f"Expected {len(meta.channels)} channels, got {c_dim}."
        )

    return meta.channels_meta


class Image(AbstractImage[ImageMetaHandler]):
    """A class to handle a single image (or level) in an OME-Zarr image.

    This class is meant to be subclassed by specific image types.
    """

    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        path: str,
        meta_handler: ImageMetaHandler | None,
    ) -> None:
        """Initialize the Image at a single level.

        Args:
            group_handler: The Zarr group handler.
            path: The path to the image in the omezarr file.
            meta_handler: The image metadata handler.

        """
        if meta_handler is None:
            meta_handler = ImplementedImageMetaHandlers().find_meta_handler(
                group_handler
            )
        super().__init__(
            group_handler=group_handler, path=path, meta_handler=meta_handler
        )
        self._channels_meta = _check_channel_meta(self.meta, self.dimensions)

    @property
    def meta(self) -> NgioImageMeta:
        """Return the metadata."""
        return self._meta_handler.meta

    @property
    def channel_labels(self) -> list[str]:
        """Return the channels of the image."""
        channel_labels = []
        for c in self._channels_meta.channels:
            channel_labels.append(c.label)
        return channel_labels

    @property
    def wavelength_ids(self) -> list[str | None]:
        """Return the list of wavelength of the image."""
        wavelength_ids = []
        for c in self._channels_meta.channels:
            wavelength_ids.append(c.wavelength_id)
        return wavelength_ids

    @property
    def num_channels(self) -> int:
        """Return the number of channels."""
        return len(self._channels_meta.channels)

    def consolidate(
        self,
        order: Literal[0, 1, 2] = 1,
        mode: Literal["dask", "numpy", "coarsen"] = "dask",
    ) -> None:
        """Consolidate the label on disk."""
        consolidate_image(self, order=order, mode=mode)


class ImagesContainer:
    """A class to handle the /labels group in an OME-NGFF file."""

    def __init__(self, group_handler: ZarrGroupHandler) -> None:
        """Initialize the LabelGroupHandler."""
        self._group_handler = group_handler
        self._meta_handler = ImplementedImageMetaHandlers().find_meta_handler(
            group_handler
        )

    @property
    def meta(self) -> NgioImageMeta:
        """Return the metadata."""
        return self._meta_handler.meta

    @property
    def levels(self) -> int:
        """Return the number of levels in the image."""
        return self._meta_handler.meta.levels

    @property
    def levels_paths(self) -> list[str]:
        """Return the paths of the levels in the image."""
        return self._meta_handler.meta.paths

    @property
    def num_channels(self) -> int:
        """Return the number of channels."""
        image = self.get()
        return image.num_channels

    @property
    def channel_labels(self) -> list[str]:
        """Return the channels of the image."""
        image = self.get()
        return image.channel_labels

    @property
    def wavelength_ids(self) -> list[str | None]:
        """Return the wavelength of the image."""
        image = self.get()
        return image.wavelength_ids

    def initialize_channel_meta(
        self,
        labels: Collection[str] | int | None = None,
        wavelength_id: Collection[str] | None = None,
        percentiles: tuple[float, float] | None = None,
        colors: Collection[str] | None = None,
        active: Collection[bool] | None = None,
        **omero_kwargs: dict,
    ) -> None:
        """Create a ChannelsMeta object with the default unit.

        Args:
            labels(Collection[str] | int): The list of channels names in the image.
                If an integer is provided, the channels will be named "channel_i".
            wavelength_id(Collection[str] | None): The wavelength ID of the channel.
                If None, the wavelength ID will be the same as the channel name.
            percentiles(tuple[float, float] | None): The start and end percentiles
                for each channel. If None, the percentiles will not be computed.
            colors(Collection[str, NgioColors] | None): The list of colors for the
                channels. If None, the colors will be random.
            active (Collection[bool] | None):active(bool): Whether the channel should
                be shown by default.
            omero_kwargs(dict): Extra fields to store in the omero attributes.
        """
        ref = self.get()

        if percentiles is not None:
            start, end = compute_image_percentile(
                ref, start_percentile=percentiles[0], end_percentile=percentiles[1]
            )
        else:
            start, end = None, None

        if labels is None:
            labels = ref.num_channels

        channel_meta = ChannelsMeta.default_init(
            labels=labels,
            wavelength_id=wavelength_id,
            colors=colors,
            start=start,
            end=end,
            active=active,
            data_type=ref.dtype,
            **omero_kwargs,
        )

        meta = self.meta
        meta.set_channels_meta(channel_meta)
        self._meta_handler.write_meta(meta)

    def update_percentiles(
        self,
        start_percentile: float = 0.1,
        end_percentile: float = 99.9,
    ) -> None:
        """Update the percentiles of the channels."""
        if self.meta._channels_meta is None:
            raise NgioValidationError("The channels meta is not initialized.")

        image = self.get()
        starts, ends = compute_image_percentile(
            image, start_percentile=start_percentile, end_percentile=end_percentile
        )

        channels = []
        for c, channel in enumerate(self.meta._channels_meta.channels):
            new_v = ChannelVisualisation(
                start=starts[c],
                end=ends[c],
                **channel.channel_visualisation.model_dump(exclude={"start", "end"}),
            )
            new_c = Channel(
                channel_visualisation=new_v,
                **channel.model_dump(exclude={"channel_visualisation"}),
            )
            channels.append(new_c)

        new_meta = ChannelsMeta(channels=channels)

        meta = self.meta
        meta.set_channels_meta(new_meta)
        self._meta_handler.write_meta(meta)

    def derive(
        self,
        store: StoreOrGroup,
        ref_path: str | None = None,
        shape: Collection[int] | None = None,
        chunks: Collection[int] | None = None,
        xy_scaling_factor: float = 2.0,
        z_scaling_factor: float = 1.0,
        overwrite: bool = False,
    ) -> "ImagesContainer":
        """Create an OME-Zarr image from a numpy array."""
        return derive_image_container(
            image_container=self,
            store=store,
            ref_path=ref_path,
            shape=shape,
            chunks=chunks,
            xy_scaling_factor=xy_scaling_factor,
            z_scaling_factor=z_scaling_factor,
            overwrite=overwrite,
        )

    def get(
        self,
        path: str | None = None,
        pixel_size: PixelSize | None = None,
        highest_resolution: bool = True,
    ) -> Image:
        """Get an image at a specific level."""
        if path is not None or pixel_size is not None:
            highest_resolution = False
        dataset = self._meta_handler.meta.get_dataset(
            path=path, pixel_size=pixel_size, highest_resolution=highest_resolution
        )
        return Image(
            group_handler=self._group_handler,
            path=dataset.path,
            meta_handler=self._meta_handler,
        )


def compute_image_percentile(
    image: Image,
    start_percentile: float = 0.1,
    end_percentile: float = 99.9,
) -> tuple[list[float], list[float]]:
    """Compute the start and end percentiles for each channel of an image.

    Args:
        image: The image to compute the percentiles for.
        start_percentile: The start percentile to compute.
        end_percentile: The end percentile to compute.

    Returns:
        A tuple containing the start and end percentiles for each channel.
    """
    starts, ends = [], []
    for c in range(image.num_channels):
        if image.num_channels == 1:
            data = image.get_array(mode="dask").ravel()
        else:
            data = image.get_array(c=c, mode="dask").ravel()
        # remove all the zeros
        mask = data > 1e-16
        data = data[mask]
        _data = data.compute()
        if _data.size == 0:
            starts.append(0.0)
            ends.append(0.0)
            continue

        # compute the percentiles
        _s_perc, _e_perc = da.percentile(
            data, [start_percentile, end_percentile], method="nearest"
        ).compute()

        starts.append(float(_s_perc))
        ends.append(float(_e_perc))

    return starts, ends


def derive_image_container(
    image_container: ImagesContainer,
    store: StoreOrGroup,
    ref_path: str | None = None,
    shape: Collection[int] | None = None,
    chunks: Collection[int] | None = None,
    xy_scaling_factor: float = 2.0,
    z_scaling_factor: float = 1.0,
    overwrite: bool = False,
) -> ImagesContainer:
    """Create an OME-Zarr image from a numpy array."""
    if ref_path is None:
        ref_image = image_container.get()
    else:
        ref_image = image_container.get(path=ref_path)

    ref_meta = ref_image.meta

    if shape is None:
        shape = ref_image.shape
    else:
        if len(shape) != len(ref_image.shape):
            raise NgioValidationError(
                "The shape of the new image does not match the reference image."
            )

    if chunks is None:
        chunks = ref_image.chunks
    else:
        if len(chunks) != len(ref_image.chunks):
            raise NgioValidationError(
                "The chunks of the new image does not match the reference image."
            )

    handler = _create_empty_image(
        store=store,
        shape=shape,
        xy_pixelsize=ref_image.pixel_size.x,
        z_spacing=ref_image.pixel_size.z,
        time_spacing=ref_image.pixel_size.t,
        levels=ref_meta.levels,
        xy_scaling_factor=xy_scaling_factor,
        z_scaling_factor=z_scaling_factor,
        time_unit=ref_image.pixel_size.time_unit,
        space_unit=ref_image.pixel_size.space_unit,
        axes_names=ref_image.dataset.axes_mapper.on_disk_axes_names,
        chunks=chunks,
        dtype=ref_image.dtype,
        overwrite=overwrite,
        version=ref_meta.version,
    )

    image_container = ImagesContainer(handler)

    if ref_image.num_channels == image_container.num_channels:
        labels = ref_image.channel_labels
        wavelength_id = ref_image.wavelength_ids
        colors = [
            c.channel_visualisation.color for c in ref_image._channels_meta.channels
        ]
        active = [
            c.channel_visualisation.active for c in ref_image._channels_meta.channels
        ]

        image_container.initialize_channel_meta(
            labels=labels,
            wavelength_id=wavelength_id,
            percentiles=None,
            colors=colors,
            active=active,
        )
    else:
        image_container.initialize_channel_meta()

    return image_container



"""Abstract class for handling OME-NGFF images."""

# %%
from collections.abc import Collection
from typing import Literal, overload

import numpy as np

from ngio.images.create import _create_empty_image
from ngio.images.image import Image, ImagesContainer
from ngio.images.label import Label, LabelsContainer
from ngio.ome_zarr_meta import (
    NgioImageMeta,
    PixelSize,
)
from ngio.ome_zarr_meta.ngio_specs import (
    SpaceUnits,
    TimeUnits,
)
from ngio.tables import (
    FeatureTable,
    MaskingROITable,
    RoiTable,
    Table,
    TablesContainer,
    TypedTable,
)
from ngio.utils import (
    AccessModeLiteral,
    NgioValidationError,
    NgioValueError,
    StoreOrGroup,
    ZarrGroupHandler,
)


def _default_table_container(handler: ZarrGroupHandler) -> TablesContainer | None:
    """Return a default table container."""
    success, table_handler = handler.safe_derive_handler("tables")
    if success and isinstance(table_handler, ZarrGroupHandler):
        return TablesContainer(table_handler)


def _default_label_container(handler: ZarrGroupHandler) -> LabelsContainer | None:
    """Return a default label container."""
    success, label_handler = handler.safe_derive_handler("labels")
    if success and isinstance(label_handler, ZarrGroupHandler):
        return LabelsContainer(label_handler)


class OmeZarrContainer:
    """This class contains an OME-Zarr image and its associated tables and labels."""

    _images_container: ImagesContainer
    _labels_container: LabelsContainer | None
    _tables_container: TablesContainer | None

    def __init__(
        self,
        store: StoreOrGroup,
        cache: bool = False,
        mode: AccessModeLiteral = "r+",
        table_container: TablesContainer | None = None,
        label_container: LabelsContainer | None = None,
        validate_arrays: bool = True,
    ) -> None:
        """Initialize the OmeZarrContainer."""
        self._group_handler = ZarrGroupHandler(store, cache, mode)
        self._images_container = ImagesContainer(self._group_handler)

        if label_container is None:
            label_container = _default_label_container(self._group_handler)
        self._labels_container = label_container

        if table_container is None:
            table_container = _default_table_container(self._group_handler)
        self._tables_container = table_container

    def __repr__(self) -> str:
        """Return a string representation of the image."""
        num_labels = len(self.list_labels())
        num_tables = len(self.list_tables())

        base_str = f"OmeZarrContainer(levels={self.levels}"
        if num_labels > 0 and num_labels < 3:
            base_str += f", labels={self.list_labels()}"
        elif num_labels >= 3:
            base_str += f", #labels={num_labels}"
        if num_tables > 0 and num_tables < 3:
            base_str += f", tables={self.list_tables()}"
        elif num_tables >= 3:
            base_str += f", #tables={num_tables}"
        base_str += ")"
        return base_str

    @property
    def images_container(self) -> ImagesContainer:
        """Return the image container."""
        return self._images_container

    @property
    def labels_container(self) -> LabelsContainer:
        """Return the labels container."""
        if self._labels_container is None:
            raise NgioValidationError("No labels found in the image.")
        return self._labels_container

    @property
    def tables_container(self) -> TablesContainer:
        """Return the tables container."""
        if self._tables_container is None:
            raise NgioValidationError("No tables found in the image.")
        return self._tables_container

    @property
    def image_meta(self) -> NgioImageMeta:
        """Return the image metadata."""
        return self._images_container.meta

    @property
    def levels(self) -> int:
        """Return the number of levels in the image."""
        return self._images_container.levels

    @property
    def levels_paths(self) -> list[str]:
        """Return the paths of the levels in the image."""
        return self._images_container.levels_paths

    def initialize_channel_meta(
        self,
        labels: Collection[str] | int | None = None,
        wavelength_id: Collection[str] | None = None,
        percentiles: tuple[float, float] | None = None,
        colors: Collection[str] | None = None,
        active: Collection[bool] | None = None,
        **omero_kwargs: dict,
    ) -> None:
        """Create a ChannelsMeta object with the default unit."""
        self._images_container.initialize_channel_meta(
            labels=labels,
            wavelength_id=wavelength_id,
            percentiles=percentiles,
            colors=colors,
            active=active,
            **omero_kwargs,
        )

    def update_percentiles(
        self,
        start_percentile: float = 0.1,
        end_percentile: float = 99.9,
    ) -> None:
        """Update the percentiles of the image."""
        self._images_container.update_percentiles(
            start_percentile=start_percentile, end_percentile=end_percentile
        )

    def get_image(
        self,
        path: str | None = None,
        pixel_size: PixelSize | None = None,
        highest_resolution: bool = True,
    ) -> Image:
        """Get an image at a specific level."""
        return self._images_container.get(
            path=path,
            pixel_size=pixel_size,
            highest_resolution=highest_resolution,
        )

    def derive_image(
        self,
        store: StoreOrGroup,
        ref_path: str | None = None,
        shape: Collection[int] | None = None,
        chunks: Collection[int] | None = None,
        xy_scaling_factor: float = 2.0,
        z_scaling_factor: float = 1.0,
        copy_tables: bool = False,
        copy_labels: bool = False,
        overwrite: bool = False,
    ) -> "OmeZarrContainer":
        """Derive a new image from the current image."""
        if copy_labels:
            raise NotImplementedError("Copying labels is not yet implemented.")

        if copy_tables:
            raise NotImplementedError("Copying tables is not yet implemented.")

        _ = self._images_container.derive(
            store=store,
            ref_path=ref_path,
            shape=shape,
            chunks=chunks,
            xy_scaling_factor=xy_scaling_factor,
            z_scaling_factor=z_scaling_factor,
            overwrite=overwrite,
        )
        return OmeZarrContainer(
            store=store,
            cache=False,
            mode="r+",
            table_container=None,
            label_container=None,
        )

    def list_tables(self) -> list[str]:
        """List all tables in the image."""
        if self._tables_container is None:
            return []
        return self._tables_container.list()

    @overload
    def get_table(self, name: str, check_type: None) -> Table: ...

    @overload
    def get_table(self, name: str, check_type: Literal["roi_table"]) -> RoiTable: ...

    @overload
    def get_table(
        self, name: str, check_type: Literal["masking_roi_table"]
    ) -> MaskingROITable: ...

    @overload
    def get_table(
        self, name: str, check_type: Literal["feature_table"]
    ) -> FeatureTable: ...

    def get_table(self, name: str, check_type: TypedTable | None = None) -> Table:
        """Get a table from the image."""
        if self._tables_container is None:
            raise NgioValidationError("No tables found in the image.")

        table = self._tables_container.get(name)
        match check_type:
            case "roi_table":
                if not isinstance(table, RoiTable):
                    raise NgioValueError(
                        f"Table '{name}' is not a ROI table. Found type: {table.type()}"
                    )
                return table
            case "masking_roi_table":
                if not isinstance(table, MaskingROITable):
                    raise NgioValueError(
                        f"Table '{name}' is not a masking ROI table. "
                        f"Found type: {table.type()}"
                    )
                return table
            case "feature_table":
                if not isinstance(table, FeatureTable):
                    raise NgioValueError(
                        f"Table '{name}' is not a feature table. "
                        f"Found type: {table.type()}"
                    )
                return table
            case None:
                return table
            case _:
                raise NgioValueError(f"Unknown check_type: {check_type}")

    def add_table(
        self,
        name: str,
        table: Table,
        backend: str | None = None,
        overwrite: bool = False,
    ) -> None:
        """Add a table to the image."""
        if self._tables_container is None:
            raise NgioValidationError("No tables found in the image.")
        self._tables_container.add(
            name=name, table=table, backend=backend, overwrite=overwrite
        )

    def list_labels(self) -> list[str]:
        """List all labels in the image."""
        if self._labels_container is None:
            return []
        return self._labels_container.list()

    def get_label(self, name: str, path: str) -> Label:
        """Get a label from the image."""
        if self._labels_container is None:
            raise NgioValidationError("No labels found in the image.")
        return self._labels_container.get(name=name, path=path)

    def derive_label(
        self,
        name: str,
        ref_image: Image | None = None,
        shape: Collection[int] | None = None,
        chunks: Collection[int] | None = None,
        dtype: str = "uint16",
        xy_scaling_factor=2.0,
        z_scaling_factor=1.0,
        overwrite: bool = False,
    ) -> Label:
        """Derive a label from an image."""
        if self._labels_container is None:
            raise NgioValidationError("No labels found in the image.")

        if ref_image is None:
            ref_image = self.get_image()
        self._labels_container.derive(
            name=name,
            ref_image=ref_image,
            shape=shape,
            chunks=chunks,
            dtype=dtype,
            xy_scaling_factor=xy_scaling_factor,
            z_scaling_factor=z_scaling_factor,
            overwrite=overwrite,
        )
        return self.get_label(name, path="0")


def open_omezarr_container(
    store: StoreOrGroup,
    cache: bool = False,
    mode: AccessModeLiteral = "r+",
    validate_arrays: bool = True,
) -> OmeZarrContainer:
    """Open an OME-Zarr image."""
    return OmeZarrContainer(
        store=store,
        cache=cache,
        mode=mode,
        validate_arrays=validate_arrays,
    )


def open_image(
    store: StoreOrGroup,
    path: str | None = None,
    pixel_size: PixelSize | None = None,
    highest_resolution: bool = False,
    cache: bool = False,
    mode: AccessModeLiteral = "r+",
) -> Image:
    """Open a single level image from an OME-Zarr image."""
    group_handler = ZarrGroupHandler(store, cache, mode)
    images_container = ImagesContainer(group_handler)
    return images_container.get(
        path=path,
        pixel_size=pixel_size,
        highest_resolution=highest_resolution,
    )


def create_empty_omezarr(
    store: StoreOrGroup,
    shape: Collection[int],
    xy_pixelsize: float,
    z_spacing: float = 1.0,
    time_spacing: float = 1.0,
    levels: int | list[str] = 5,
    xy_scaling_factor: float = 2,
    z_scaling_factor: float = 1.0,
    space_unit: SpaceUnits | str | None = None,
    time_unit: TimeUnits | str | None = None,
    axes_names: Collection[str] | None = None,
    name: str | None = None,
    chunks: Collection[int] | None = None,
    dtype: str = "uint16",
    channel_labels: list[str] | None = None,
    channel_wavelengths: list[str] | None = None,
    percentiles: tuple[float, float] | None = None,
    channel_colors: Collection[str] | None = None,
    channel_active: Collection[bool] | None = None,
    overwrite: bool = False,
    version: str = "0.4",
) -> OmeZarrContainer:
    """Create an empty OME-Zarr image with the given shape and metadata.

    Args:
        store (StoreOrGroup): The Zarr store or group to create the image in.
        shape (Collection[int]): The shape of the image.
        xy_pixelsize (float): The pixel size in x and y dimensions.
        z_spacing (float, optional): The spacing between z slices. Defaults to 1.0.
        time_spacing (float, optional): The spacing between time points.
            Defaults to 1.0.
        levels (int | list[str], optional): The number of levels in the pyramid or a
            list of level names. Defaults to 5.
        xy_scaling_factor (float, optional): The down-scaling factor in x and y
            dimensions. Defaults to 2.0.
        z_scaling_factor (float, optional): The down-scaling factor in z dimension.
            Defaults to 1.0.
        space_unit (SpaceUnits | str | None, optional): The unit of space. Defaults to
            None.
        time_unit (TimeUnits | str | None, optional): The unit of time. Defaults to
            None.
        axes_names (Collection[str] | None, optional): The names of the axes.
            If None the canonical names are used. Defaults to None.
        name (str | None, optional): The name of the image. Defaults to None.
        chunks (Collection[int] | None, optional): The chunk shape. If None the shape
            is used. Defaults to None.
        dtype (str, optional): The data type of the image. Defaults to "uint16".
        channel_labels (list[str] | None, optional): The labels of the channels.
            Defaults to None.
        channel_wavelengths (list[str] | None, optional): The wavelengths of the
            channels. Defaults to None.
        percentiles (tuple[float, float] | None, optional): The percentiles of the
            channels. Defaults to None.
        channel_colors (Collection[str] | None, optional): The colors of the channels.
            Defaults to None.
        channel_active (Collection[bool] | None, optional): Whether the channels are
            active. Defaults to None.
        overwrite (bool, optional): Whether to overwrite an existing image.
            Defaults to True.
        version (str, optional): The version of the OME-Zarr specification.
            Defaults to "0.4".
    """
    handler = _create_empty_image(
        store=store,
        shape=shape,
        xy_pixelsize=xy_pixelsize,
        z_spacing=z_spacing,
        time_spacing=time_spacing,
        levels=levels,
        xy_scaling_factor=xy_scaling_factor,
        z_scaling_factor=z_scaling_factor,
        space_unit=space_unit,
        time_unit=time_unit,
        axes_names=axes_names,
        name=name,
        chunks=chunks,
        dtype=dtype,
        overwrite=overwrite,
        version=version,
    )

    omezarr = OmeZarrContainer(store=handler.store, mode="r+")
    omezarr.initialize_channel_meta(
        labels=channel_labels,
        wavelength_id=channel_wavelengths,
        percentiles=percentiles,
        colors=channel_colors,
        active=channel_active,
    )
    return omezarr


def create_omezarr_from_array(
    store: StoreOrGroup,
    array: np.ndarray,
    xy_pixelsize: float,
    z_spacing: float = 1.0,
    time_spacing: float = 1.0,
    levels: int | list[str] = 5,
    xy_scaling_factor: float = 2.0,
    z_scaling_factor: float = 1.0,
    space_unit: SpaceUnits | str | None = None,
    time_unit: TimeUnits | str | None = None,
    axes_names: Collection[str] | None = None,
    channel_labels: list[str] | None = None,
    channel_wavelengths: list[str] | None = None,
    percentiles: tuple[float, float] | None = (0.1, 99.9),
    channel_colors: Collection[str] | None = None,
    channel_active: Collection[bool] | None = None,
    name: str | None = None,
    chunks: Collection[int] | None = None,
    overwrite: bool = False,
    version: str = "0.4",
) -> OmeZarrContainer:
    """Create an OME-Zarr image from a numpy array.

    Args:
        store (StoreOrGroup): The Zarr store or group to create the image in.
        array (np.ndarray): The image data.
        xy_pixelsize (float): The pixel size in x and y dimensions.
        z_spacing (float, optional): The spacing between z slices. Defaults to 1.0.
        time_spacing (float, optional): The spacing between time points.
            Defaults to 1.0.
        levels (int | list[str], optional): The number of levels in the pyramid or a
            list of level names. Defaults to 5.
        xy_scaling_factor (float, optional): The down-scaling factor in x and y
            dimensions. Defaults to 2.0.
        z_scaling_factor (float, optional): The down-scaling factor in z dimension.
            Defaults to 1.0.
        space_unit (SpaceUnits | str | None, optional): The unit of space. Defaults to
            None.
        time_unit (TimeUnits | str | None, optional): The unit of time. Defaults to
            None.
        axes_names (Collection[str] | None, optional): The names of the axes.
            If None the canonical names are used. Defaults to None.
        name (str | None, optional): The name of the image. Defaults to None.
        chunks (Collection[int] | None, optional): The chunk shape. If None the shape
            is used. Defaults to None.
        channel_labels (list[str] | None, optional): The labels of the channels.
            Defaults to None.
        channel_wavelengths (list[str] | None, optional): The wavelengths of the
            channels. Defaults to None.
        percentiles (tuple[float, float] | None, optional): The percentiles of the
            channels. Defaults to None.
        channel_colors (Collection[str] | None, optional): The colors of the channels.
            Defaults to None.
        channel_active (Collection[bool] | None, optional): Whether the channels are
            active. Defaults to None.
        overwrite (bool, optional): Whether to overwrite an existing image.
            Defaults to True.
        version (str, optional): The version of the OME-Zarr specification.
            Defaults to "0.4".
    """
    handler = _create_empty_image(
        store=store,
        shape=array.shape,
        xy_pixelsize=xy_pixelsize,
        z_spacing=z_spacing,
        time_spacing=time_spacing,
        levels=levels,
        xy_scaling_factor=xy_scaling_factor,
        z_scaling_factor=z_scaling_factor,
        space_unit=space_unit,
        time_unit=time_unit,
        axes_names=axes_names,
        name=name,
        chunks=chunks,
        dtype=array.dtype,
        overwrite=overwrite,
        version=version,
    )

    omezarr = OmeZarrContainer(store=handler.store, mode="r+")
    image = omezarr.get_image()
    image.set_array(array)
    image.consolidate()
    omezarr.initialize_channel_meta(
        labels=channel_labels,
        wavelength_id=channel_wavelengths,
        percentiles=percentiles,
        colors=channel_colors,
        active=channel_active,
    )
    return omezarr



"""Utilities for reading and writing OME-Zarr metadata."""

from ngio.ome_zarr_meta._generic_handlers import (
    BaseImageMetaHandler,
    BaseLabelMetaHandler,
    ImageMetaHandler,
    LabelMetaHandler,
)
from ngio.ome_zarr_meta._meta_handlers import (
    ImplementedImageMetaHandlers,
    ImplementedLabelMetaHandlers,
    open_image_meta_handler,
)
from ngio.ome_zarr_meta.ngio_specs import (
    AxesMapper,
    Dataset,
    NgioImageMeta,
    NgioLabelMeta,
    PixelSize,
)

__all__ = [
    "AxesMapper",
    "BaseImageMetaHandler",
    "BaseLabelMetaHandler",
    "Dataset",
    "ImageMetaHandler",
    "ImplementedImageMetaHandlers",
    "ImplementedLabelMetaHandlers",
    "LabelMetaHandler",
    "NgioImageMeta",
    "NgioLabelMeta",
    "PixelSize",
    "open_image_meta_handler",
]



"""A module for handling label images in OME-NGFF files."""

from collections.abc import Collection
from typing import Literal

from ngio.images.abstract_image import AbstractImage, consolidate_image
from ngio.images.create import _create_empty_label
from ngio.images.image import Image
from ngio.ome_zarr_meta import (
    ImplementedLabelMetaHandlers,
    LabelMetaHandler,
    NgioLabelMeta,
)
from ngio.ome_zarr_meta.ngio_specs import SpaceUnits, TimeUnits
from ngio.utils import (
    NgioValidationError,
    NgioValueError,
    StoreOrGroup,
    ZarrGroupHandler,
)


class Label(AbstractImage[LabelMetaHandler]):
    """Placeholder class for a label."""

    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        path: str,
        meta_handler: LabelMetaHandler | None,
    ) -> None:
        """Initialize the Image at a single level.

        Args:
            group_handler: The Zarr group handler.
            path: The path to the image in the omezarr file.
            meta_handler: The image metadata handler.

        """
        if meta_handler is None:
            meta_handler = ImplementedLabelMetaHandlers().find_meta_handler(
                group_handler
            )
        super().__init__(
            group_handler=group_handler, path=path, meta_handler=meta_handler
        )

    @property
    def meta(self) -> NgioLabelMeta:
        """Return the metadata."""
        return self._meta_handler.meta

    def consolidate(
        self,
        mode: Literal["dask", "numpy", "coarsen"] = "dask",
    ) -> None:
        """Consolidate the label on disk."""
        consolidate_image(self, mode=mode, order=0)


class LabelsContainer:
    """A class to handle the /labels group in an OME-NGFF file."""

    def __init__(self, group_handler: ZarrGroupHandler) -> None:
        """Initialize the LabelGroupHandler."""
        self._group_handler = group_handler

        # Validate the group
        # Either contains a labels attribute or is empty
        attrs = self._group_handler.load_attrs()
        if len(attrs) == 0:
            # It's an empty group
            pass
        elif "labels" in attrs and isinstance(attrs["labels"], list):
            # It's a valid group
            pass
        else:
            raise NgioValidationError(
                f"Invalid /labels group. "
                f"Expected a single labels attribute with a list of label names. "
                f"Found: {attrs}"
            )

    def list(self) -> list[str]:
        """Create the /labels group if it doesn't exist."""
        attrs = self._group_handler.load_attrs()
        return attrs.get("labels", [])

    def get(self, name: str, path: str) -> Label:
        """Get a label from the group."""
        group_handler = self._group_handler.derive_handler(name)
        return Label(group_handler, path, None)

    def derive(
        self,
        name: str,
        ref_image: Image,
        shape: Collection[int] | None = None,
        chunks: Collection[int] | None = None,
        dtype: str = "uint16",
        xy_scaling_factor=2.0,
        z_scaling_factor=1.0,
        overwrite: bool = False,
    ) -> None:
        """Add a label to the group."""
        existing_labels = self.list()
        if name in existing_labels and not overwrite:
            raise NgioValueError(
                f"Table '{name}' already exists in the group. "
                "Use overwrite=True to replace it."
            )

        label_group = self._group_handler.get_group(name, create_mode=True)

        _derive_label(
            ref_image=ref_image,
            store=label_group,
            shape=shape,
            chunks=chunks,
            dtype=dtype,
            xy_scaling_factor=xy_scaling_factor,
            z_scaling_factor=z_scaling_factor,
            overwrite=overwrite,
        )

        if name not in existing_labels:
            existing_labels.append(name)
            self._group_handler.write_attrs({"labels": existing_labels})

    def new(
        self,
        name: str,
        shape: Collection[int],
        xy_pixelsize: float,
        z_spacing: float = 1.0,
        time_spacing: float = 1.0,
        levels: "int | list[str]" = 5,
        xy_scaling_factor: float = 2.0,
        z_scaling_factor: float = 1.0,
        space_unit: SpaceUnits | str | None = None,
        time_unit: TimeUnits | str | None = None,
        axes_names: Collection[str] | None = None,
        chunks: Collection[int] | None = None,
        dtype: str = "uint16",
        overwrite: bool = False,
        version: str = "0.4",
    ) -> None:
        """Add a label to the group."""
        existing_labels = self.list()
        if name in existing_labels and not overwrite:
            raise NgioValueError(
                f"Table '{name}' already exists in the group. "
                "Use overwrite=True to replace it."
            )

        label_group = self._group_handler.get_group(name, create_mode=True)

        _create_empty_label(
            store=label_group,
            shape=shape,
            xy_pixelsize=xy_pixelsize,
            z_spacing=z_spacing,
            time_spacing=time_spacing,
            levels=levels,
            xy_scaling_factor=xy_scaling_factor,
            z_scaling_factor=z_scaling_factor,
            space_unit=space_unit,
            time_unit=time_unit,
            axes_names=axes_names,
            chunks=chunks,
            dtype=dtype,
            overwrite=overwrite,
            version=version,
        )

        if name not in existing_labels:
            existing_labels.append(name)
            self._group_handler.write_attrs({"labels": existing_labels})


def _derive_label(
    ref_image: Image,
    store: StoreOrGroup,
    shape: Collection[int] | None = None,
    chunks: Collection[int] | None = None,
    dtype: str = "uint16",
    xy_scaling_factor=2.0,
    z_scaling_factor=1.0,
    overwrite: bool = False,
) -> None:
    """Create an OME-Zarr image from a numpy array."""
    ref_meta = ref_image.meta
    # remove channls if present
    shape_ref = ref_image.shape
    chunks_ref = ref_image.chunks
    axes_names_ref = ref_image.dataset.axes_mapper.on_disk_axes_names
    c_axis = ref_image.dataset.axes_mapper.get_index("c")
    if c_axis is not None:
        shape_ref = shape_ref[:c_axis] + shape_ref[c_axis + 1 :]
        chunks_ref = chunks_ref[:c_axis] + chunks_ref[c_axis + 1 :]
        axes_names_ref = axes_names_ref[:c_axis] + axes_names_ref[c_axis + 1 :]

    if shape is None:
        shape = shape_ref

    if chunks is None:
        chunks = chunks_ref

    if len(shape) != len(shape_ref):
        raise NgioValidationError(
            "The shape of the new image does not match the reference image."
        )

    if len(chunks) != len(chunks_ref):
        raise NgioValidationError(
            "The chunks of the new image does not match the reference image."
        )

    _ = _create_empty_label(
        store=store,
        shape=shape,
        xy_pixelsize=ref_image.pixel_size.x,
        z_spacing=ref_image.pixel_size.z,
        time_spacing=ref_image.pixel_size.t,
        levels=ref_meta.levels,
        xy_scaling_factor=xy_scaling_factor,
        z_scaling_factor=z_scaling_factor,
        time_unit=ref_image.pixel_size.time_unit,
        space_unit=ref_image.pixel_size.space_unit,
        axes_names=axes_names_ref,
        chunks=chunks,
        dtype=dtype,
        overwrite=overwrite,
        version=ref_meta.version,
    )
    return None



"""Base class for handling OME-NGFF metadata in Zarr groups."""

from typing import Generic, Protocol, TypeVar

from pydantic import ValidationError

from ngio.ome_zarr_meta.ngio_specs import AxesSetup, NgioImageMeta, NgioLabelMeta
from ngio.utils import (
    NgioValueError,
    ZarrGroupHandler,
)

ConverterError = ValidationError | Exception | None


class ImageMetaHandler(Protocol):
    """Protocol for OME-Zarr image handlers."""

    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = True,
    ):
        """Initialize the handler."""
        ...

    def safe_load_meta(self) -> NgioImageMeta | ConverterError:
        """Load the metadata from the store."""
        ...

    @property
    def meta(self) -> NgioImageMeta:
        """Return the metadata."""
        ...

    def write_meta(self, meta: NgioImageMeta) -> None:
        """Write the metadata to the store."""
        ...


class LabelMetaHandler(Protocol):
    """Protocol for OME-Zarr label handlers."""

    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = True,
    ):
        """Initialize the handler."""
        ...

    def safe_load_meta(self) -> NgioLabelMeta | ConverterError:
        """Load the metadata from the store."""
        ...

    @property
    def meta(self) -> NgioLabelMeta:
        """Return the metadata."""
        ...

    def write_meta(self, meta: NgioLabelMeta) -> None:
        """Write the metadata to the store."""
        ...


###########################################################################
#
# The code below implements a generic class for handling OME-Zarr metadata
# in Zarr groups.
#
###########################################################################


class ImageMetaImporter(Protocol):
    @staticmethod
    def __call__(
        metadata: dict,
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = True,
    ) -> tuple[bool, NgioImageMeta | ConverterError]:
        """Convert the metadata to a NgioImageMeta object.

        Args:
            metadata (dict): The metadata (typically from a Zarr group .attrs).
            axes_setup (AxesSetup, optional): The axes setup.
                This is used to map axes with non-canonical names.
            allow_non_canonical_axes (bool, optional): Whether to allow non-canonical
                axes.
            strict_canonical_order (bool, optional): Whether to enforce a strict
                canonical order.

        Returns:
            tuple[bool, NgioImageMeta | ConverterError]: A tuple with a boolean
                indicating whether the conversion was successful and the
                NgioImageMeta object or an error.

        """
        ...


class ImageMetaExporter(Protocol):
    def __call__(self, metadata: NgioImageMeta) -> dict: ...


class LabelMetaImporter(Protocol):
    @staticmethod
    def __call__(
        metadata: dict,
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = True,
    ) -> tuple[bool, NgioLabelMeta | ConverterError]:
        """Convert the metadata to a NgioLabelMeta object.

        Args:
            metadata (dict): The metadata (typically from a Zarr group .attrs).
            axes_setup (AxesSetup, optional): The axes setup.
                This is used to map axes with non-canonical names.
            allow_non_canonical_axes (bool, optional): Whether to allow non-canonical
                axes.
            strict_canonical_order (bool, optional): Whether to enforce a strict
                canonical order.

        Returns:
            tuple[bool, NgioLabelMeta | ConverterError]: A tuple with a boolean
                indicating whether the conversion was successful and the
                NgioLabelMeta object or an error.

        """
        ...


class LabelMetaExporter(Protocol):
    def __call__(self, metadata: NgioLabelMeta) -> dict: ...


_meta = TypeVar("_meta", NgioImageMeta, NgioLabelMeta)
_meta_importer = TypeVar("_meta_importer", ImageMetaImporter, LabelMetaImporter)
_meta_exporter = TypeVar("_meta_exporter", ImageMetaExporter, LabelMetaExporter)


class GenericMetaHandler(Generic[_meta, _meta_importer, _meta_exporter]):
    """Generic class for handling OME-Zarr metadata in Zarr groups."""

    def __init__(
        self,
        meta_importer: _meta_importer,
        meta_exporter: _meta_exporter,
        group_handler: ZarrGroupHandler,
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = True,
    ):
        """Initialize the handler.

        Args:
            meta_importer (MetaImporter): The metadata importer.
            meta_exporter (MetaExporter): The metadata exporter.
            group_handler (ZarrGroupHandler): The Zarr group handler.
            axes_setup (AxesSetup, optional): The axes setup.
                This is used to map axes with non-canonical names.
            allow_non_canonical_axes (bool, optional): Whether to allow non-canonical
                axes.
            strict_canonical_order (bool, optional): Whether to enforce a strict
                canonical order.
        """
        self._group_handler = group_handler
        self._meta_importer = meta_importer
        self._meta_exporter = meta_exporter
        self._axes_setup = axes_setup
        self._allow_non_canonical_axes = allow_non_canonical_axes
        self._strict_canonical_order = strict_canonical_order

    def _load_meta(self, return_error: bool = False):
        """Load the metadata from the store."""
        attrs = self._group_handler.load_attrs()
        is_valid, meta_or_error = self._meta_importer(
            metadata=attrs,
            axes_setup=self._axes_setup,
            allow_non_canonical_axes=self._allow_non_canonical_axes,
            strict_canonical_order=self._strict_canonical_order,
        )
        if is_valid:
            return meta_or_error

        if return_error:
            return meta_or_error

        raise NgioValueError(f"Could not load metadata: {meta_or_error}")

    def safe_load_meta(self) -> _meta | ConverterError:
        """Load the metadata from the store."""
        return self._load_meta(return_error=True)

    def _write_meta(self, meta) -> None:
        """Write the metadata to the store."""
        v04_meta = self._meta_exporter(metadata=meta)
        self._group_handler.write_attrs(v04_meta)

    def write_meta(self, meta: _meta) -> None:
        """Write the metadata to the store."""
        raise NotImplementedError

    @property
    def meta(self) -> _meta:
        """Return the metadata."""
        raise NotImplementedError


class BaseImageMetaHandler(
    GenericMetaHandler[NgioImageMeta, ImageMetaImporter, ImageMetaExporter]
):
    """Generic class for handling OME-Zarr metadata in Zarr groups."""

    def __init__(
        self,
        meta_importer: ImageMetaImporter,
        meta_exporter: ImageMetaExporter,
        group_handler: ZarrGroupHandler,
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = True,
    ):
        """Initialize the handler.

        Args:
            meta_importer (ImageMetaImporter): The metadata importer.
            meta_exporter (ImageMetaExporter): The metadata exporter.
            group_handler (ZarrGroupHandler): The Zarr group handler.
            axes_setup (AxesSetup, optional): The axes setup.
                This is used to map axes with non-canonical names.
            allow_non_canonical_axes (bool, optional): Whether to allow non-canonical
                axes.
            strict_canonical_order (bool, optional): Whether to enforce a strict
                canonical order.
        """
        super().__init__(
            meta_importer=meta_importer,
            meta_exporter=meta_exporter,
            group_handler=group_handler,
            axes_setup=axes_setup,
            allow_non_canonical_axes=allow_non_canonical_axes,
            strict_canonical_order=strict_canonical_order,
        )

    def safe_load_meta(
        self, return_error: bool = False
    ) -> NgioImageMeta | ConverterError:
        """Load the metadata from the store."""
        return self._load_meta(return_error)

    @property
    def meta(self) -> NgioImageMeta:
        """Load the metadata from the store."""
        meta = self._load_meta()
        if isinstance(meta, NgioImageMeta):
            return meta
        raise NgioValueError(f"Could not load metadata: {meta}")

    def write_meta(self, meta: NgioImageMeta) -> None:
        self._write_meta(meta)


class BaseLabelMetaHandler(
    GenericMetaHandler[NgioLabelMeta, LabelMetaImporter, LabelMetaExporter]
):
    """Generic class for handling OME-Zarr metadata in Zarr groups."""

    def __init__(
        self,
        meta_importer: LabelMetaImporter,
        meta_exporter: LabelMetaExporter,
        group_handler: ZarrGroupHandler,
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = True,
    ):
        """Initialize the handler.

        Args:
            meta_importer (LabelMetaImporter): The metadata importer.
            meta_exporter (LabelMetaExporter): The metadata exporter.
            group_handler (ZarrGroupHandler): The Zarr group handler.
            axes_setup (AxesSetup, optional): The axes setup.
                This is used to map axes with non-canonical names.
            allow_non_canonical_axes (bool, optional): Whether to allow non-canonical
                axes.
            strict_canonical_order (bool, optional): Whether to enforce a strict
                canonical order.
        """
        super().__init__(
            meta_importer=meta_importer,
            meta_exporter=meta_exporter,
            group_handler=group_handler,
            axes_setup=axes_setup,
            allow_non_canonical_axes=allow_non_canonical_axes,
            strict_canonical_order=strict_canonical_order,
        )

    def safe_load_meta(
        self, return_error: bool = False
    ) -> NgioLabelMeta | ConverterError:
        """Load the metadata from the store."""
        return self._load_meta(return_error)

    @property
    def meta(self) -> NgioLabelMeta:
        """Load the metadata from the store."""
        meta = self._load_meta()
        if isinstance(meta, NgioLabelMeta):
            return meta
        raise NgioValueError(f"Could not load metadata: {meta}")

    def write_meta(self, meta: NgioLabelMeta) -> None:
        self._write_meta(meta)



"""Utility to read/write OME-Zarr metadata v0.4."""

from ngio.ome_zarr_meta.v04._meta_handlers import (
    V04ImageMetaHandler,
    V04LabelMetaHandler,
)

__all__ = [
    "V04ImageMetaHandler",
    "V04LabelMetaHandler",
]



from collections.abc import Iterable
from typing import Protocol

import pandas as pd
import pandas.api.types as ptypes

from ngio.utils import (
    NgioTableValidationError,
)


class TableValidator(Protocol):
    def __call__(self, table: pd.DataFrame) -> pd.DataFrame:
        """Validate the table DataFrame.

        A Validator is just a simple callable that takes a
            DataFrame and returns a DataFrame.

        If the DataFrame is valid, the same DataFrame is returned.
        If the DataFrame is invalid, the Validator can either modify the DataFrame
            to make it valid or raise a NgioTableValidationError.

        Args:
            table (pd.DataFrame): The DataFrame to validate.

        Returns:
            pd.DataFrame: The validated DataFrame.

        """
        ...


def validate_table(
    table_df: pd.DataFrame,
    validators: Iterable[TableValidator] | None = None,
) -> pd.DataFrame:
    """Validate the table DataFrame.

    Args:
        table_df (pd.DataFrame): The DataFrame to validate.
        validators (Collection[Validator] | None): A collection of functions
            used to validate the table. Default is None.

    Returns:
        pd.DataFrame: The validated DataFrame.
    """
    validators = validators or []

    # Apply all provided validators
    for validator in validators:
        table_df = validator(table_df)

    return table_df


####################################################################################################
#
# Common table validators
#
####################################################################################################
def validate_index_key(
    dataframe: pd.DataFrame, index_key: str | None, overwrite: bool = False
) -> pd.DataFrame:
    """Correctly set the index of the DataFrame.

    This function checks if the index_key is present in the DataFrame.
    If not it tries to set sensible defaults.

    In order:
        - If index_key is None, nothing can be done.
        - If index_key is already the index of the DataFrame, nothing is done.
        - If index_key is in the columns, we set the index to that column.
        - If current index is None, we set the index to the index_key.
        - If current index is not None and overwrite is True,
            we set the index to the index_key.

    """
    if index_key is None:
        # Nothing to do
        return dataframe

    if dataframe.index.name == index_key:
        # Index is already set to index_key correctly
        return dataframe

    if index_key in dataframe.columns:
        dataframe = dataframe.set_index(index_key)
        return dataframe

    if dataframe.index.name is None:
        dataframe.index.name = index_key
        return dataframe

    elif overwrite:
        dataframe.index.name = index_key
        return dataframe
    else:
        raise NgioTableValidationError(
            f"Index key {index_key} not found in DataFrame. "
            f"Current index is {dataframe.index.name}. If you want to overwrite the "
            "index set overwrite=True."
        )


def validate_index_dtype(dataframe: pd.DataFrame, index_type: str) -> pd.DataFrame:
    """Check if the index of the DataFrame has the correct dtype."""
    match index_type:
        case "str":
            if ptypes.is_integer_dtype(dataframe.index):
                # Convert the int index to string is generally safe
                dataframe = dataframe.set_index(dataframe.index.astype(str))

            if not ptypes.is_string_dtype(dataframe.index):
                raise NgioTableValidationError(
                    f"Table index must be of string type, got {dataframe.index.dtype}"
                )

        case "int":
            if ptypes.is_string_dtype(dataframe.index):
                # Try to convert the string index to int
                try:
                    dataframe = dataframe.set_index(dataframe.index.astype(int))
                except ValueError as e:
                    if "invalid literal for int() with base 10" in str(e):
                        raise NgioTableValidationError(
                            "Table index must be of integer type, got str."
                            f" We tried implicit conversion and failed: {e}"
                        ) from None
                    else:
                        raise e from e

            if not ptypes.is_integer_dtype(dataframe.index):
                raise NgioTableValidationError(
                    f"Table index must be of integer type, got {dataframe.index.dtype}"
                )
        case _:
            raise ValueError(f"index_type {index_type} not recognized")

    return dataframe


def validate_columns(
    table_df: pd.DataFrame,
    required_columns: list[str],
    optional_columns: list[str] | None = None,
) -> pd.DataFrame:
    """Validate the columns headers of the table.

    If a required column is missing, a TableValidationError is raised.
    If a list of optional columns is provided, only required and optional columns are
        allowed in the table.

    Args:
        table_df (pd.DataFrame): The DataFrame to validate.
        required_columns (list[str]): A list of required columns.
        optional_columns (list[str] | None): A list of optional columns.
            Default is None.

    Returns:
        pd.DataFrame: The validated DataFrame.
    """
    table_header = table_df.columns
    for column in required_columns:
        if column not in table_header:
            raise NgioTableValidationError(
                f"Could not find required column: {column} in the table"
            )

    if optional_columns is None:
        return table_df

    possible_columns = [*required_columns, *optional_columns]
    for column in table_header:
        if column not in possible_columns:
            raise NgioTableValidationError(
                f"Could not find column: {column} in the list of possible columns. ",
                f"Possible columns are: {possible_columns}",
            )

    return table_df


def validate_unique_index(table_df: pd.DataFrame) -> pd.DataFrame:
    """Validate that the index of the table is unique."""
    if table_df.index.is_unique:
        return table_df

    # Find the duplicates
    duplicates = table_df.index[table_df.index.duplicated()].tolist()
    raise NgioTableValidationError(
        f"Index of the table contains duplicates values. Duplicate: {duplicates}"
    )



"""Protocol for table backends handlers."""

from collections.abc import Collection
from typing import Literal, Protocol

from anndata import AnnData
from pandas import DataFrame

from ngio.tables.backends._anndata_v1 import AnnDataBackend
from ngio.tables.backends._json_v1 import JsonTableBackend
from ngio.utils import NgioValueError, ZarrGroupHandler


class TableBackendProtocol(Protocol):
    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        index_key: str | None = None,
        index_type: Literal["int", "str"] = "int",
    ): ...

    @staticmethod
    def backend_name() -> str: ...

    @staticmethod
    def implements_anndata() -> bool: ...

    @staticmethod
    def implements_dataframe() -> bool: ...

    def load_columns(self) -> list[str]: ...

    def load_as_anndata(self, columns: Collection[str] | None = None) -> AnnData: ...

    def load_as_dataframe(
        self, columns: Collection[str] | None = None
    ) -> DataFrame: ...

    def write_from_dataframe(
        self, table: DataFrame, metadata: dict | None = None
    ) -> None: ...

    def write_from_anndata(
        self, table: AnnData, metadata: dict | None = None
    ) -> None: ...


class ImplementedTableBackends:
    """A class to manage the available table backends."""

    _instance = None
    _implemented_backends: dict[str, type[TableBackendProtocol]]

    def __new__(cls):
        """Create a new instance of the class if it does not exist."""
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._implemented_backends = {}
        return cls._instance

    @property
    def available_backends(self) -> list[str]:
        """Return the available table backends."""
        return list(self._implemented_backends.keys())

    def get_backend(
        self,
        backend_name: str | None,
        group_handler: ZarrGroupHandler,
        index_key: str | None = None,
        index_type: Literal["int", "str"] = "int",
    ) -> TableBackendProtocol:
        """Try to get a handler for the given store based on the metadata version."""
        if backend_name is None:
            # Default to anndata since it is currently
            # the only backend in use.
            backend_name = "anndata_v1"

        if backend_name not in self._implemented_backends:
            raise NgioValueError(f"Table backend {backend_name} not implemented.")
        handler = self._implemented_backends[backend_name](
            group_handler=group_handler, index_key=index_key, index_type=index_type
        )
        return handler

    def add_backend(
        self,
        table_beckend: type[TableBackendProtocol],
        overwrite: bool = False,
    ):
        """Register a new handler."""
        backend_name = table_beckend.backend_name()
        if backend_name in self._implemented_backends and not overwrite:
            raise NgioValueError(
                f"Table backend {backend_name} already implemented. "
                "Use the `overwrite=True` parameter to overwrite it."
            )
        self._implemented_backends[backend_name] = table_beckend


ImplementedTableBackends().add_backend(AnnDataBackend)
ImplementedTableBackends().add_backend(JsonTableBackend)



"""Concrete implementation of the OME-Zarr metadata handlers for version 0.4."""

from ngio.ome_zarr_meta._generic_handlers import (
    BaseImageMetaHandler,
    BaseLabelMetaHandler,
)
from ngio.ome_zarr_meta.ngio_specs import AxesSetup
from ngio.ome_zarr_meta.v04._v04_spec_utils import (
    ngio_to_v04_image_meta,
    ngio_to_v04_label_meta,
    v04_to_ngio_image_meta,
    v04_to_ngio_label_meta,
)
from ngio.utils import ZarrGroupHandler


class V04ImageMetaHandler(BaseImageMetaHandler):
    """Base class for handling OME-Zarr 0.4 metadata."""

    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = True,
    ):
        super().__init__(
            meta_importer=v04_to_ngio_image_meta,
            meta_exporter=ngio_to_v04_image_meta,
            group_handler=group_handler,
            axes_setup=axes_setup,
            allow_non_canonical_axes=allow_non_canonical_axes,
            strict_canonical_order=strict_canonical_order,
        )


class V04LabelMetaHandler(BaseLabelMetaHandler):
    """Base class for handling OME-Zarr 0.4 metadata."""

    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = True,
    ):
        super().__init__(
            meta_importer=v04_to_ngio_label_meta,
            meta_exporter=ngio_to_v04_label_meta,
            group_handler=group_handler,
            axes_setup=axes_setup,
            allow_non_canonical_axes=allow_non_canonical_axes,
            strict_canonical_order=strict_canonical_order,
        )



"""Implementation of the Masking ROI table class.

This class follows the roi_table specification at:
https://fractal-analytics-platform.github.io/fractal-tasks-core/tables/
"""

from collections.abc import Iterable
from typing import Literal

from pydantic import BaseModel

from ngio.common import WorldCooROI
from ngio.tables._validators import validate_columns
from ngio.tables.backends import ImplementedTableBackends
from ngio.tables.v1._roi_table import (
    OPTIONAL_COLUMNS,
    REQUIRED_COLUMNS,
    _dataframe_to_rois,
    _rois_to_dataframe,
)
from ngio.utils import NgioValueError, ZarrGroupHandler


class RegionMeta(BaseModel):
    """Metadata for the region."""

    path: str


class MaskingROITableV1Meta(BaseModel):
    """Metadata for the ROI table."""

    fractal_table_version: Literal["1"] = "1"
    type: Literal["masking_roi_table"] = "masking_roi_table"
    backend: str | None = None
    region: RegionMeta | None = None
    instance_key: str = "label"


class MaskingROITableV1:
    """Class to handle fractal ROI tables.

    To know more about the ROI table format, please refer to the
    specification at:
    https://fractal-analytics-platform.github.io/fractal-tasks-core/tables/
    """

    def __init__(
        self,
        rois: Iterable[WorldCooROI] | None = None,
        reference_label: str | None = None,
    ) -> None:
        """Create a new ROI table."""
        if reference_label is None:
            self._meta = MaskingROITableV1Meta()
        else:
            path = f"../labels/{reference_label}"
            self._meta = MaskingROITableV1Meta(region=RegionMeta(path=path))
        self._table_backend = None

        self._rois = {}
        if rois is not None:
            self.add(rois)

    @staticmethod
    def type() -> Literal["masking_roi_table"]:
        """Return the type of the table."""
        return "masking_roi_table"

    @staticmethod
    def version() -> Literal["1"]:
        """Return the version of the fractal table."""
        return "1"

    @property
    def backend_name(self) -> str | None:
        """Return the name of the backend."""
        if self._table_backend is None:
            return None
        return self._table_backend.backend_name()

    @classmethod
    def _from_handler(
        cls, handler: ZarrGroupHandler, backend_name: str | None = None
    ) -> "MaskingROITableV1":
        """Create a new ROI table from a Zarr store."""
        meta = MaskingROITableV1Meta(**handler.load_attrs())

        if backend_name is None:
            backend = ImplementedTableBackends().get_backend(
                backend_name=meta.backend,
                group_handler=handler,
                index_key="label",
                index_type="int",
            )
        else:
            backend = ImplementedTableBackends().get_backend(
                backend_name=backend_name,
                group_handler=handler,
                index_key="label",
                index_type="int",
            )
            meta.backend = backend_name

        if not backend.implements_dataframe:
            raise NgioValueError(
                "The backend does not implement the dataframe protocol."
            )

        table = cls()
        table._meta = meta
        table._table_backend = backend

        dataframe = backend.load_as_dataframe()
        dataframe = validate_columns(
            dataframe,
            required_columns=REQUIRED_COLUMNS,
            optional_columns=OPTIONAL_COLUMNS,
        )
        table._rois = _dataframe_to_rois(dataframe)
        return table

    def _set_backend(
        self,
        handler: ZarrGroupHandler,
        backend_name: str | None = None,
    ) -> None:
        """Set the backend of the table."""
        backend = ImplementedTableBackends().get_backend(
            backend_name=backend_name,
            group_handler=handler,
            index_key="label",
            index_type="int",
        )
        self._meta.backend = backend_name
        self._table_backend = backend

    def rois(self) -> list[WorldCooROI]:
        """List all ROIs in the table."""
        return list(self._rois.values())

    def get(self, label: int) -> WorldCooROI:
        """Get an ROI from the table."""
        _label = str(label)
        if _label not in self._rois:
            raise KeyError(f"ROI {_label} not found in the table.")
        return self._rois[_label]

    def add(self, roi: WorldCooROI | Iterable[WorldCooROI]) -> None:
        """Append ROIs to the current table."""
        if isinstance(roi, WorldCooROI):
            roi = [roi]

        for _roi in roi:
            if _roi.name in self._rois:
                raise NgioValueError(f"ROI {_roi.name} already exists in the table.")
            self._rois[_roi.name] = _roi

    def consolidate(self) -> None:
        """Write the current state of the table to the Zarr file."""
        if self._table_backend is None:
            raise NgioValueError(
                "No backend set for the table. "
                "Please add the table to a OME-Zarr Image before calling consolidate."
            )

        dataframe = _rois_to_dataframe(self._rois, index_key="label")
        dataframe = validate_columns(
            dataframe,
            required_columns=REQUIRED_COLUMNS,
            optional_columns=OPTIONAL_COLUMNS,
        )
        self._table_backend.write_from_dataframe(
            dataframe, metadata=self._meta.model_dump(exclude_none=True)
        )



from pathlib import Path

import pandas as pd
import pytest

from ngio.tables.tables_container import open_table, write_table
from ngio.tables.v1 import GenericTable


@pytest.mark.parametrize("backend", ["json_v1", "anndata_v1"])
def test_generic_table(tmp_path: Path, backend: str):
    store = tmp_path / "test.zarr"
    test_df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    table = GenericTable(test_df)

    write_table(store=store, table=table, backend=backend)

    loaded_table = open_table(store=store)
    assert isinstance(loaded_table, GenericTable)
    assert set(loaded_table.dataframe.columns) == {"a", "b"}
    for column in loaded_table.dataframe.columns:
        pd.testing.assert_series_equal(
            loaded_table.dataframe[column], test_df[column], check_index=False
        )



"""Region of interest (ROI) metadata.

These are the interfaces bwteen the ROI tables / masking ROI tables and
    the ImageLikeHandler.
"""

import numpy as np
from pydantic import BaseModel, ConfigDict, Field

from ngio.common._dimensions import Dimensions
from ngio.ome_zarr_meta.ngio_specs import PixelSize, SpaceUnits


def _to_raster(value: float, pixel_size: float, max_shape: int) -> int:
    """Convert to raster coordinates."""
    round_value = int(np.round(value / pixel_size))
    # Ensure the value is within the image shape boundaries
    return max(0, min(round_value, max_shape))


def _to_world(value: int, pixel_size: float) -> float:
    """Convert to world coordinates."""
    return value * pixel_size


class WorldCooROI(BaseModel):
    """Region of interest (ROI) metadata."""

    name: str
    x_length: float
    y_length: float
    z_length: float = 1.0
    x: float = 0.0
    y: float = 0.0
    z: float = 0.0
    unit: SpaceUnits = Field(SpaceUnits.micrometer, repr=False)

    model_config = ConfigDict(extra="allow")

    def to_raster_coo(
        self, pixel_size: PixelSize, dimensions: Dimensions
    ) -> "RasterCooROI":
        """Convert to raster coordinates."""
        dim_x = dimensions.get("x")
        dim_y = dimensions.get("y")
        # Will default to 1 if z does not exist
        dim_z = dimensions.get("z", strict=False)

        return RasterCooROI(
            name=self.name,
            x=_to_raster(self.x, pixel_size.x, dim_x),
            y=_to_raster(self.y, pixel_size.y, dim_y),
            z=_to_raster(self.z, pixel_size.z, dim_z),
            x_length=_to_raster(self.x_length, pixel_size.x, dim_x),
            y_length=_to_raster(self.y_length, pixel_size.y, dim_y),
            z_length=_to_raster(self.z_length, pixel_size.z, dim_z),
        )


class RasterCooROI(BaseModel):
    """Region of interest (ROI) metadata."""

    name: str
    x: int
    y: int
    z: int
    x_length: int
    y_length: int
    z_length: int
    model_config = ConfigDict(extra="allow")

    def to_world_coo_roi(self, pixel_size: PixelSize) -> WorldCooROI:
        """Convert to world coordinates."""
        return WorldCooROI(
            name=self.name,
            x=_to_world(self.x, pixel_size.x),
            y=_to_world(self.y, pixel_size.y),
            z=_to_world(self.z, pixel_size.z),
            x_length=_to_world(self.x_length, pixel_size.x),
            y_length=_to_world(self.y_length, pixel_size.y),
            z_length=_to_world(self.z_length, pixel_size.z),
            unit=pixel_size.space_unit,
        )

    def to_slices(self) -> dict[str, slice]:
        """Return the slices for the ROI."""
        return {
            "x": slice(self.x, self.x + self.x_length),
            "y": slice(self.y, self.y + self.y_length),
            "z": slice(self.z, self.z + self.z_length),
        }



from pathlib import Path

import fsspec
import numpy as np
import pytest

from ngio import create_empty_omezarr, create_omezarr_from_array, open_omezarr_container


@pytest.mark.parametrize("array_mode", ["numpy", "dask"])
def test_omezarr_container(tmp_path: Path, array_mode: str):
    # Very basic test to check if the container is working
    # to be expanded with more meaningful tests
    store = tmp_path / "omezarr.zarr"
    omezarr = create_empty_omezarr(
        store,
        shape=(10, 20, 30),
        chunks=(1, 20, 30),
        xy_pixelsize=0.5,
        levels=3,
        dtype="uint8",
    )

    assert isinstance(omezarr.__repr__(), str)
    assert omezarr.levels == 3
    assert omezarr.levels_paths == ["0", "1", "2"]

    image = omezarr.get_image()
    assert image.shape == (10, 20, 30)
    assert image.dtype == "uint8"
    assert image.chunks == (1, 20, 30)
    assert image.pixel_size.x == 0.5

    array = image.get_array(
        x=slice(None), axes_order=["c", "z", "y", "x"], mode=array_mode
    )

    assert array.shape == (1, 10, 20, 30)

    array = array + 1

    image.set_array(array, x=slice(None), axes_order=["c", "z", "y", "x"])
    image.consolidate(mode=array_mode)

    # Omemeta
    omezarr.initialize_channel_meta(labels=["channel_x"])
    image = omezarr.get_image()
    assert image.channel_labels == ["channel_x"]
    omezarr.update_percentiles()

    image = omezarr.get_image(path="2")
    assert np.mean(image.get_array()) == 1

    new_omezarr = omezarr.derive_image(tmp_path / "derived.zarr", ref_path="2")

    assert new_omezarr.levels == 3
    new_image = new_omezarr.get_image()
    assert new_image.shape == image.shape


def test_create_omezarr_container(tmp_path: Path):
    # Very basic test to check if the container is working
    # to be expanded with more meaningful tests
    store = tmp_path / "omezarr.zarr"
    create_omezarr_from_array(
        store,
        array=np.zeros((10, 20, 30), dtype="uint8"),
        xy_pixelsize=0.5,
        levels=3,
    )


def test_remote_omezarr_container():
    url = (
        "https://raw.githubusercontent.com/"
        "fractal-analytics-platform/fractal-ome-zarr-examples/"
        "refs/heads/main/v04/"
        "20200812-CardiomyocyteDifferentiation14-Cycle1_B_03_mip.zarr/"
    )

    fs = fsspec.implementations.http.HTTPFileSystem(client_kwargs={})
    store = fs.get_mapper(url)

    omezarr = open_omezarr_container(store)

    assert omezarr.list_labels() == ["nuclei"]
    assert omezarr.list_tables() == [
        "FOV_ROI_table",
        "nuclei_ROI_table",
        "well_ROI_table",
        "regionprops_DAPI",
    ]

    _ = omezarr.get_label("nuclei", path="0")
    _ = omezarr.get_table("well_ROI_table")
